/*	check.c

	This program is designed to check the TTL_PAPERS hardware for
	correct operation.  It does not offer particularly good
	diagnostics because the design of TTL_PAPERS makes it very
	difficult to debug hardware problems -- there isn't any way to
	distinguish many of the possible failure modes.

	Hardware problems to watch for:

	1.	Is your wiring color code correct for all your cables?
		Even cables obtained in a single shipment from the
		same source sometimes will employ different color
		codes for a few, but not all, of the wires.

	2.	Inconsistent wiring of PEs to the logic.  Generally,
		connections are not sensitive to which PE is wired to
		which pin within a group, but a few signals are.  E.g.,
		the P_SEL signal and P_RDY must be paired correctly.

	3.	The most common trouble we've had is rooted in a bad
		power supply.  If the hardware seems completely lost,
		check the AC voltage across the board's +5 and ground.
		If you don't have the right capacitors filtering the
		input, and you have an AC-to-DC adapter that isn't very
		good, you might find as much as a few volts worth of
		AC ripple.  The +5 connection should be between +4.9
		and +5.1 with virtually no AC ripple if things are
		done right -- even if you have a really lousy adapter.

	This program tests the hardware assuming that ONLY this one PE
	is physically connected to the unit.  TTL_PAPERS was designed
	to take advantage of the fact that TTL inputs float high, so
	most of the basic functionality can be checked by testing one
	PE connection at a time.

	Notice that this code can be used to test the VAPERS stuff as
	well, but I'm not sure what purpose that serves....

	1995 by Hank Dietz
*/


#include <stdio.h>
#include "paperslib.h"
#include "intern.h"


#define	P_MODE(x) \
	outb(((unsigned char)(x)), ((unsigned short) P_PORTBASE+2))
#define	P_NAK	0x01		/* Not interrupt ack */
#define	P_IRQ	0x02		/* Interrupt request */
#define	P_SEL	0x04		/* Int/Rdy Select */


double	outspeed = 0.0;
double	inspeed = 0.0;
double	iospeed = 0.0;


yes()
{
	int c;

	printf("Type n<ret> for no, <ret> for yes: ");
	c = getchar();
	if (c != '\n') {
		while (getchar() != '\n');
	}
	return(c != 'n');
}


chk(char * s)
{
	printf("%s\n", s);
	printf("Do you want to resume the testing?\n");

	if (!yes()) {
		/* Make the LED orange (yellow) again... */
		P_OUT(P_LY);

		exit(3);
	}
}

page()
{
	putchar('\n');
}

main(register int argc,
register char **argv,
register char **arge)
{


#ifdef	P_VAPERS
	register int live = 0;

#define	LOOPCOUNT	100

	/* Check-in with the simulator... */
	p_init();

	/* Pretend to disconnect all but one PE... */
	if (argc > 2) {
		fprintf(stderr, "Usage: %s [PE_to_check]\n");
		exit(1);
	}
	if (argc == 2) {
		live = (atoi(argv[1]) & (NPROC-1));
	}

	if (IPROC != live) {
		P_OUT(0xff);
		P_MODE(0xff);		/* Not really right... */
		return(0);
	}

	printf("Although this program was designed to test the TTL_PAPERS\n");
	printf("hardware one PE-connection at a time, it can be used to\n");
	printf("perform similar tests on the simulated TTL_VAPERS hardware.\n");
	printf("Although you don't need to have the TTL_VAPERS unit displayed\n");
	printf("graphically, you'll probably want to have xvapers running so\n");
	printf("that you can watch the LEDs change color.  The particular\n");
	printf("simulated PE connection being tested is specified on the\n");
	printf("command line to this program, or is PE0 by default.  For this\n");
	printf("run, you have selected PE%d.  The other three PEs will simply\n",
	       live);
	printf("act as though they have no connection to the simulated unit.\n");
	page();


#else	/* ifndef P_VAPERS */


#define	LOOPCOUNT	100000

	/* Parse command line... */
	if (argc != 1) {
		fprintf(stderr, "Usage: %s\n");
		exit(1);
	}

	/* Get port access... */
	if (ioperm(P_PORTBASE, 3, 1)) {
		fprintf(stderr,
			"%s: port access refused (need to be priv)\n");
		exit(2);
	}

	printf("To use this program, a TTL_PAPERS unit should be connected to\n");
	printf("this machine's parallel printer port at 0x%x.  Only this one\n",
	       P_PORTBASE);
	printf("machine should be connected to the TTL_PAPERS unit during this\n");
	printf("check; the other inputs to TTL_PAPERS should simply float high.\n");
	printf("Because I/O lines are floating, and floating lines are highly\n");
	printf("susceptible to noise, this program may detect errors even\n");
	printf("though the hardware functions perfectly when fully connected\n");
	printf("to four machines.  However, if the hardware passes this test,\n");
	printf("there is a very high probability that it is correct.  Of course,\n");
	printf("you should run this test 4 times, once for each PE connection.\n");
	page();


#endif


power:
	printf("The blue power LED on the TTL_PAPERS front panel should be\n");
	printf("lit.  Is it?\n");
	if (!yes()) {
		printf("Are any of the TTL_PAPERS LEDs lit?\n");
		if (yes()) {
			chk("Check the blue LED and its connections.\n");
		} else {
			chk("Check power supply and check board for +5/gnd shorts.\n");
		}
		goto power;
	}
	page();

	/* All LEDs orange/yellow? */
	P_OUT(P_LY);
	P_MODE(P_IRQ | P_SEL | 0x08);
	printf("All four bi-color LEDs on the TTL_PAPERS front panel display\n");
	printf("should now be orange (or perhaps yellow).  Are they?\n");
	if (!yes()) {
		chk("Your TTL_PAPERS unit has a wiring error.\n");
		goto power;
	}
	page();

cable:
	/* Proper cable connection? */
	P_OUT(P_LB);
	printf("Three of the four bi-color LEDs on the TTL_PAPERS front panel\n");
	printf("display should now be orange (yellow); the one black bi-color\n");
	printf("LED corresponds to this PE.  Is this what you see?\n");
	if (!yes()) {
		printf("Are all the LEDs still orange (yellow)?\n");
		if (!yes()) {
			chk("Your TTL_PAPERS unit has a wiring error.\n");
		} else {
			chk("Check cable connection and port number.\n");
		}
		goto power;
	}
	page();

	/* Check the lights individually */
	P_OUT(P_LG);
	printf("Is this PE's LED now green?\n");
	if (!yes()) {
		printf("Is this PE's LED now red?\n");
		if (yes()) {
			chk("You probably confused the red and green leads of this LED.\n");
		} else {
			chk("This LED is not wired correctly.\n");
		}
		goto power;
	}
	page();

	P_OUT(P_LR);
	printf("Is this PE's LED now red?\n");
	if (!yes()) {
		chk("This LED is not wired correctly.\n");
		goto power;
	}
	page();

	/* Turn off the interrupt stuff... */
	P_MODE(0);
	P_MODE(0);
	P_MODE(P_NAK);
	P_MODE(P_NAK);

	/* Time port access */
	printf("Now timing port I/O speed....\n");
	{
		register int i;

		i = LOOPCOUNT;
		p_timerStart();
		do {
			P_OUT(P_LG);
			P_OUT(P_LG);
			P_OUT(P_LG);
			P_OUT(P_LG);
			P_OUT(P_LG);
			P_OUT(P_LG);
			P_OUT(P_LG);
			P_OUT(P_LG);
			P_OUT(P_LG);
			P_OUT(P_LG);
		} while (--i);
		outspeed = (100000 / LOOPCOUNT) * p_timerElapsed();
		printf("Typical port output speed is 1-4us, this is %gus.\n",
		       outspeed);

		i = LOOPCOUNT;
		p_timerStart();
		do {
			P_IN();
			P_IN();
			P_IN();
			P_IN();
			P_IN();
			P_IN();
			P_IN();
			P_IN();
			P_IN();
			P_IN();
		} while (--i);
		inspeed = (100000 / LOOPCOUNT) * p_timerElapsed();
		printf("Typical port input speed is 1-4us, this is %gus.\n",
		       inspeed);

		i = LOOPCOUNT;
		p_timerStart();
		do {
			P_OUT(P_LG);
			P_IN();
			P_OUT(P_LG);
			P_IN();
			P_OUT(P_LG);
			P_IN();
			P_OUT(P_LG);
			P_IN();
			P_OUT(P_LG);
			P_IN();
		} while (--i);
		iospeed = (100000 / LOOPCOUNT) * p_timerElapsed();
		printf("Typical mixed output/input speed is 1-4us, this is %gus.\n",
		       iospeed);
	}
	page();

	/* Now the tricky stuff... */

	printf("Now checking barrier synchronization...\n");
	P_OUT(P_LG);
	{
		register int i = LOOPCOUNT;
		do {
			if (barchk() == 1) goto power;
		} while (--i);
	}
	printf("Barrier synchronization logic is ok.\n");
	page();

	/* Ok, set-up just enough so p_nand() works... */
	P_OUT(last_out = (P_S0 | P_LG));
	P_OUT(last_out);
	P_OUT(last_out);
	P_OUT(last_out);
	P_OUT(last_out);
	P_OUT(last_out);
	P_OUT(last_out);
	P_OUT(last_out);
	P_OUT(last_out);
	P_OUT(last_out);

	printf("Now checking data NANDing...\n");
	P_OUT(P_LG);
	{
		register int i = (10 * LOOPCOUNT);
		do {
			register int x = ((rand() / 13) & 0xf);
			register int y, z;

			if (((y = p_nand(x ^ 0xf)) != x) ||
			    ((z = p_nand(x ^ 0xf)) != x)) {
				printf("Data error: sent 0x%x, got 0x%x, then 0x%x\n",
				       x, y, z);
				if (z == x) {
					printf("This data glitch was probably due to noise that would not be\n");
					printf("present in a fully connected system.  You should perform\n");
					printf("additional tests with the TTL_PAPERS unit connected to four\n");
					printf("machines to confirm/disprove this hypothesis.\n");
				} else {
					chk("Check wiring of P_D? and P_I?, AC noise on +5/gnd, and cables.\n");
					page();

					printf("Continuing data NANDing...\n");
				}
			}
		} while (--i);
	}
	printf("Data NANDing logic is ok.\n");
	page();

	printf("This PE connection's portion of your TTL_PAPERS unit appears\n");
	printf("to be functional.  Be sure to test the other PE connections.\n");

	/* Make the LED orange (yellow) again... */
	P_OUT(P_LY);

	/* Success */
	exit(0);
}

barchk()
{
	/* This code doesn't use the normal barrier synchronization
	   logic because it needs to be able to detect specific error
	   conditions that would not be observable using that code.
	*/

	P_OUT(P_S0 | P_LG);
	if ((!(P_IN() & P_RDY)) && (!(P_IN() & P_RDY))) {
		if ((!(P_IN() & P_RDY)) && (!(P_IN() & P_RDY))) {
			chk("Check wiring of P_S0, P_SEL, P_RDY\n");
		} else {
			register int i;

			for (i=0; i<100000; ++i) {
				if (!(P_IN() & P_RDY)) {
					chk("P_RDY is unstable; check the power supply for AC noise.\n");
					return(1);
				}
			}

			printf("It took more than one port operation for the P_RDY line to\n");
			printf("settle appropriately.  Either your port is too fast or your\n");
			printf("cables are too long for the TTL_PAPERS support as written.\n");
			if ((outspeed > 0.5) && (inspeed > 0.5)) {
				chk("Your cables seem to be too long.\n");
			} else {
				chk("Port access is unusually fast on this machine.  You might want\nto modify the inline.h routines by inserting extra port ops to\ncompensate for the port speed, but shortening your cables could\nyield significantly higher performance.\n");
			}
		}
		return(1);
	}
	P_OUT(P_S1 | P_LG);
	if ((P_IN() & P_RDY) && (P_IN() & P_RDY)) {
		if ((P_IN() & P_RDY) && (P_IN() & P_RDY)) {
			chk("Check wiring of P_S0, P_SEL, P_RDY\n");
		} else {
			register int i;

			for (i=0; i<100000; ++i) {
				if (P_IN() & P_RDY) {
					chk("P_RDY is unstable; check the power supply for AC noise.\n");
					return(1);
				}
			}

			printf("It took more than one port operation for the P_RDY line to\n");
			printf("settle appropriately.  Either your port is too fast or your\n");
			printf("cables are too long for the TTL_PAPERS support as written.\n");
			if ((outspeed > 0.5) && (inspeed > 0.5)) {
				chk("Your cables seem to be too long.\n");
			} else {
				chk("Port access is unusually fast on this machine.  You might want\nto modify the inline.h routines by inserting extra port ops to\ncompensate for the port speed, but shortening your cables could\nyield significantly higher performance.\n");
			}
		}
		return(1);
	}

	return(0);
}
