/*	gather.c

	PAPERS gather operations.  These operations are the core
	mechanism for reductions, etc.

	Gather data from all PEs.  Our contribution is datum.  The
	gathered values from all PEs are placed in the consecutive
	locations starting at dest, i.e., PEk's datum gets placed in
	*(dest + k).  Values are changed only for PEs that participate
	in the barriers (PEs in mask).

	1995 by Hank Dietz
*/


#include "papers.h"
#include "gather.h"
#include "intern.h"


#ifdef	P_VAPERS
static void
#else
extern inline void
#endif
p_put8u(register uint8 datum)
{
	p_nand(datum & 0xf);
	p_nand((datum >> 4) & 0xf);
}

#ifdef	P_VAPERS
static uint8
#else
extern inline uint8
#endif
p_get8u(void)
{
	register uint8 datum = (0xff ^ p_nand(0xf));
	return(datum ^ (p_nand(0xf) << 4));
}

#ifdef	P_VAPERS
static uint8
#else
extern inline void
#endif
p_void8u(void)
{
	/* Like doing a p_get8u() and ignoring the value */
	p_nand(0xf);
	p_nand(0xf);
}


/*	Basic object gather operations...
*/
#define	GATHER(name, type, bitcnt) \
void \
name(register type *dest, \
register type datum) \
{ \
	static type d; \
	register int who; \
 \
	d = datum; \
	for (who=0; who<NPROC; ++who) { \
		if ((1 << who) & last_mask) { \
			if (IPROC == who) { \
				/* Our turn to send and store data */ \
				register uint8 *p = ((uint8 *) &d); \
				register uint8 *e = (p + (bitcnt / 8)); \
				do { \
					p_put8u(*p); \
				} while (++p < e); \
				*(dest + who) = d; \
			} else { \
				/* We just get and store data... */ \
				register uint8 *p = ((uint8 *) (dest + who)); \
				register uint8 *e = (p + (bitcnt / 8)); \
				do { \
					*p = p_get8u(); \
				} while (++p < e); \
			} \
		} \
	} \
}


GATHER(p_gather8,    int8,  8)
GATHER(p_gather8u,  uint8,  8)
GATHER(p_gather16,   int16, 16)
GATHER(p_gather16u, uint16, 16)
GATHER(p_gather32,   int32, 32)
GATHER(p_gather32u, uint32, 32)
GATHER(p_gather64,   int64, 64)
GATHER(p_gather64u, uint64, 64)
GATHER(p_gatherf,      f32, 32)
GATHER(p_gatherd,      f64, 64)


void
p_gather1u(register uint1 *dest,
register uint1 datum)
{
	/* Assumes 32-bit ints... */
	p_gatherBits32u(dest, (datum != 0), 1);
}


/*	Bit sequence gather operations...
*/
#define	GATHERBITS(name, type) \
void \
name(register type *dest, \
register type datum, \
register int bits) \
{ \
	/* If we're not the only PE in our mask, work to do...  */ \
	if (last_mask & ~OUR_MASK) { \
		register barrier mask = last_mask; \
		register type bit = 1; \
		register type t0 = 0; \
		register type t1 = 0; \
		register type t2 = 0; \
		register type t3 = 0; \
 \
		/* For each bit... */ \
		while (--bits >= 0) { \
			/* Get a 4-bit stripe of data */ \
			register barrier x = \
				p_waitvec(datum & bit); \
 \
			/* Convert bits... */ \
			if (x & 1) t0 |= bit; \
			if (x & 2) t1 |= bit; \
			if (x & 4) t2 |= bit; \
			if (x & 8) t3 |= bit; \
 \
			/* Next bit... */ \
			bit <<= 1; \
		} \
 \
		/* Now that we've got them, where do we put them? */ \
		if (mask & 1) { \
			*(dest + 0) = t0; \
		} \
		if (mask & 2) { \
			*(dest + 1) = t1; \
		} \
		if (mask & 4) { \
			*(dest + 2) = t2; \
		} \
		if (mask & 8) { \
			*(dest + 3) = t3; \
		} \
	} else {\
		/* Just store our value...  or bits of it */ \
		*(dest + IPROC) = (((1 << bits) - 1) & datum); \
	} \
 \
}


GATHERBITS(p_gatherBits8u,  uint8)
GATHERBITS(p_gatherBits16u, uint16)
GATHERBITS(p_gatherBits32u, uint32)
GATHERBITS(p_gatherBits64u, uint64)
