/*	papers.c

	Support for basic functions using PAPERS.

	1995 by Hank Dietz
*/


#include <stdio.h>
#include <stdlib.h>
#include "version.h"


/*	Include the internal support header, which includes the inline
	library, but set the DO_NOT_INLINE flag to force the library to
	be expanded as ordinary function definitions rather than inline
	expansions.  This is the only place where the flag is set, and
	simply ensures that linkable versions of the functions will be
	created in papers.o.
*/
#define DO_NOT_INLINE
#include "intern.h"


barrier last_mask = 0;			/* Last barrier mask */
portdata last_out = (P_S0 | P_LG);	/* Last output value */

int iproc = -1;			/* PAPERS PE number */
barrier our_mask;		/* Our barrier mask */
int ixproc;			/* X coordinate */
int iyproc;			/* Y coordinate */


#ifdef	P_VAPERS


/*	Stuff for pipe communication between the TTL_VAPERS simulator
	and the PE processes that it creates...
*/
static int	p_pipes[NPROC+1][2];	/* Pipes */
static int	p_pids[NPROC];		/* PE PIDs */


int		p_req_pipe;		/* Request pipe */
int		p_res_pipe;		/* Response pipe */


/*	Stuff for tracking port status
*/
static uint8 reg[3][NPROC] = {
	0xff, 0xff, 0xff, 0xff,
	0x07, 0x07, 0x07, 0x07,
	0x01, 0x01, 0x01, 0x01 };

static int waiting[NPROC] = { 0, 0, 0, 0 };
static int ready[NPROC] = { -1, -1, -1, -1 };


static int p_earlycount[NPROC] = { 0, 0, 0, 0 };
static int p_barcount = 0;


/*	Status file stuff...
*/
static int p_statfd = -1;
static char statfile[512];
static char p_statbuf[(NPROC+1) * 8 * 8];
static int p_dirty = 0;
static unsigned int p_oldval[] = {
	-1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1,
	-1, -1, -1, -1, -1, -1, -1, -1, -1,
};

#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>


static volatile int stopatbar = 0;

static void
sig1(int a)
{
	signal(SIGUSR1, &sig1);
	++stopatbar;
}

static void
sig2(int a)
{
	signal(SIGUSR2, &sig2);
	stopatbar = 0;
}


static void
p_setstat(register int pos,
register unsigned int val)
{
	/* Conversion to 8-char hex ASCII in status buffer...
	   but only do it if the file exists and the data value
	   we're putting there differs from what was there.
	*/
	if ((p_statfd != -1) &&
	    (p_oldval[pos] != val)) {
		register char *p = &(p_statbuf[(pos) * 8]);
		register char *q = (p + 7);

		p_oldval[pos] = val;

		do {
			*q = "0123456789abcdef"[val & 15];
			val >>= 4;
			} while ((q--) > p);

		p_dirty = 1;
	}
}

#define	p_upstat() \
	{ \
		if (p_dirty) { \
			lseek(p_statfd, 0, SEEK_SET); \
			write(p_statfd, &(p_statbuf[0]), sizeof(p_statbuf)); \
			p_dirty = 0; \
		} \
	}

static void
p_statfile()
{
	register char *p;
	register int i;

	/* Make and open the status file... */
	p = getenv("PAPERS_STATFILE");
	if (p == NULL) {
		p = "VAPERS_status";
	}

	strncpy(statfile, p, 511);
	statfile[512] = 0;

	p_statfd = open(statfile, (O_RDWR | O_CREAT), 0600);
	if (p_statfd == -1) {
		fprintf(stderr,
			"VAPERS:  cannot create/open PAPERS_STATFILE \"%s\"\n",
			statfile);
		return;
	}

	/* Format the status for each PE...
	*/
	for (i=0; i<NPROC; ++i) {
		p_setstat((i * 8), p_pids[i]);
		p_setstat(((i * 8) + 1), reg[0][i]);
		p_setstat(((i * 8) + 2), reg[1][i]);
		p_setstat(((i * 8) + 3), reg[2][i]);
		p_setstat(((i * 8) + 4), p_earlycount[i]);
		p_setstat(((i * 8) + 5), 0);
		p_setstat(((i * 8) + 6), 0);
		p_setstat(((i * 8) + 7), 0);
	}

	/* The status for the system as a whole...
	*/
	p_setstat((NPROC * 8), getpid());
	p_setstat(((NPROC * 8) + 1), p_barcount);
	p_setstat(((NPROC * 8) + 2), 0);
	p_setstat(((NPROC * 8) + 3), 0);
	p_setstat(((NPROC * 8) + 4), 0);
	p_setstat(((NPROC * 8) + 5), 0);
	p_setstat(((NPROC * 8) + 6), 0);
	p_setstat(((NPROC * 8) + 7), 0);


	/* Make the file look like the above status...
	*/
	p_upstat();
}


static void
p_simulate()
{
	p_portmesg mesg;
	register int rdy = 0;
	register int irq = 0;


	/* Process port interactions just like TTL_PAPERS would
	*/
	while (read(p_pipes[NPROC][0], &mesg, sizeof(mesg)) ==
	       sizeof(mesg)) {
		register int from = mesg.from;
		register int port = mesg.port;
		register int value = mesg.value;
		uint8 tmp;

		switch (mesg.op) {
		case P_INB:
			/* Reply with current status */
			tmp = reg[port][from];
			write(p_pipes[from][1], &tmp, sizeof(uint8));
			if (port == 1) {
				ready[from] = rdy;
			}
			break;
		case P_WAITRDY:
			/* Reply with current status when RDY changed */
			if (ready[from] != rdy) {
				tmp = reg[1][from];
				write(p_pipes[from][1], &tmp, sizeof(uint8));
				ready[from] = rdy;
			} else {
				waiting[from] = 1;
			}

			/* In any case, we got here early... */
			++(p_earlycount[from]);
			p_setstat(((from * 8) + 4), p_earlycount[from]);

			break;
		case P_OUTB:
			/* Change status to reflect new port output */
			if (mesg.port == 0) {
				register int i, anded, ibits;

				reg[ 0 ][ mesg.from ] = mesg.value;
				p_setstat(((from * 8) + 1), mesg.value);

				anded = reg[ 0 ][ 0 ];
				for (i=1; i<NPROC; ++i) {
					anded &= reg[ 0 ][ i ];
				}

				ibits = (((0xf & ~anded) << 3) | 0x7);

				switch (anded & (P_S1 | P_S0)) {
				case P_S1:
					if (rdy) {
						while (stopatbar == 1) pause();
						if (stopatbar > 1) stopatbar = 1;
						++p_barcount;
						p_setstat(((NPROC * 8) + 1), p_barcount);
					}
					rdy = 0;
					break;
				case P_S0:
					if (!rdy) {
						while (stopatbar == 1) pause();
						if (stopatbar > 1) stopatbar = 1;
						++p_barcount;
						p_setstat(((NPROC * 8) + 1), p_barcount);
					}
					rdy = P_RDY;
					break;
				}

				for (i=0; i<NPROC; ++i) {
					reg[1][i] = (ibits | rdy);
					p_setstat(((i * 8) + 2), reg[1][i]);
				}

				for (i=0; i<NPROC; ++i) {
					if (waiting[i] &&
					    (ready[i] != rdy)) {
						tmp = reg[1][i];
						write(p_pipes[i][1], &tmp, sizeof(uint8));
						ready[i] = rdy;
						waiting[i] = 0;
					}
				}
			}
		}

		/* Update the stat file, if appropriate */
		p_upstat();
	}
}

static void
p_hw_startup()
{
	register int i;

	/* Set-up barrier stop signals...
	*/
	signal(SIGUSR1, &sig1);
	signal(SIGUSR2, &sig2);

	last_out = (P_S1 | P_LG | P_D3 | P_D2 | P_D1 | P_D0);
	for (i=0; i<NPROC; ++i) {
		reg[0][i] = last_out;
		reg[1][i] = 0x7;
	}

	pipe( &(p_pipes[NPROC][0]) );

	for (iproc=0; iproc<NPROC; ++iproc) {
		pipe( &(p_pipes[iproc][0]) );
		if ((p_pids[iproc] = fork()) == 0) {
			/* Just created process iproc...
			   close unwanted pipes and return
			*/
			for (i=0; i<iproc; ++i) {
				close(p_pipes[i][1]);
			}
			close(p_pipes[iproc][1]);
			close(p_pipes[NPROC][0]);
			p_req_pipe = p_pipes[NPROC][1];
			p_res_pipe = p_pipes[iproc][0];

			our_mask = (1 << IPROC);		/* Our barrier mask */
			ixproc = (IPROC & (NXPROC - 1));	/* X coordinate */
			iyproc = (IPROC >> LNXPROC);		/* Y coordinate */

			/* Enqueue global barrier and synchronize... */
			p_enqueue(ALL_MASK);
			return;
		} else {
			/* We are the parent...
			   close the child end of the pipe
			*/
			close(p_pipes[iproc][0]);
		}
	}

	/* We are the parent...
	   close the input end of our pipe
	*/
	close(p_pipes[NPROC][1]);
	iproc = -1;

	/* Set-up status file */
	p_statfile();

	/* Now, simulate TTL_PAPERS */
	p_simulate();

	/* Kill the children before we quit */
	for (i=0; i<NPROC; ++i) {
		kill(p_pids[i], SIGKILL);
	}

	/* Get rid of the stat file... */
	if (p_statfd != -1) {
		close(p_statfd);
		unlink(statfile);
	}

	exit(0);
}


#else	/* ifndef P_VAPERS */


static primes[12] = {
	1, 2, 3, 5, 7, 11, 13, 17, 19, 23, 27, 31
};


static int
p_hw_startup(void)
{
	/* This routine uses a whole bunch of raw syncs to try to ensure
	   that everything starts together.  It may also be used to ensure
	   that synchronization is maintained within a program.  Returns
	   only once sync has been obtained, and returns the number of
	   retries required to obtain sync (0 if all PEs were in sync).
	*/
	register int tries = -1;
	register int i, j;

	our_mask = (1 << IPROC);		/* Our barrier mask */
	ixproc = (IPROC & (NXPROC - 1));	/* X coordinate */
	iyproc = (IPROC >> LNXPROC);		/* Y coordinate */

	/* Enqueue global barrier and synchronize... */
	p_enqueue(ALL_MASK);

restart:
	/* Bump counter and report problems, if any... */
	if (++tries == 1000000) {
		fprintf(stderr,
			"\nPAPERS PE%d:  after %d tries, initial sync not obtained...\n",
			IPROC, tries);
		fprintf(stderr,
			"PAPERS PE%d:  are all PEs properly connected?\n",
			IPROC);
		fprintf(stderr,
			"PAPERS PE%d:  are environment values consistent across all PEs?\n",
			IPROC);
	}

	/* Wait for everybody to send 1 bits... (disconnected) */
	P_OUT(last_out = (P_S1 | P_LB | P_D3 | P_D2 | P_D1 | P_D0));
	while (P_IN() & (P_RDY | P_I3 | P_I2 | P_I1 | P_I0));

	P_OUT(last_out = (P_S1 | P_LB | P_D3 | P_D2 | P_D1));
	do {
		i = (P_IN() & (P_RDY | P_I3 | P_I2 | P_I1 | P_I0));
	} while ((i & (P_RDY | P_I3 | P_I2 | P_I1 | P_I0)) == 0);
	if ((i & (P_RDY | P_I3 | P_I2 | P_I1 | P_I0)) != P_I0) {
		goto restart;
	}

	P_OUT(last_out = (P_S1 | P_LB | P_D3 | P_D2));
	do {
		i = (P_IN() & (P_RDY | P_I3 | P_I2 | P_I1 | P_I0));
	} while ((i & (P_RDY | P_I3 | P_I2 | P_I1 | P_I0)) == P_I0);
	if ((i & (P_RDY | P_I3 | P_I2 | P_I1 | P_I0)) != (P_I0 | P_I1)) {
		goto restart;
	}

	usleep(10000);
	P_OUT(last_out = (P_S1 | P_LB | P_D3 | P_D2 | P_D1 | P_D0));
	while (P_IN() & (P_RDY | P_I3 | P_I2 | P_I1 | P_I0));
	last_out |= P_LG;
	while (p_nand(0xf) != 0) {
	}

	/* Now go through prime-number sequence of each PE flipping
	   the one bit that corresponds to its IPROC...  this is a
	   very simple sequence, but should yield a sync with a very
	   high probability....
	*/
	{
		register int i, j;

		for (i=0; i<12; ++i) {
			for (j=primes[i]; j>0; --j) {
				if (p_waitvec(1) != ALL_MASK) {
					goto restart;
				}
			}
			if (p_waitvec(0) != 0) {
				goto restart;
			}
		}
	}

	/* Ok, return number of retries */
	return(tries);
}


#endif


static void
p_initenv(void)
{
	/* Initialize ONLY the environment-sensitive info... */
	register char *p;

	/* Check environment variables */
	p = getenv("PAPERS");
	if ((p == NULL) || strcmp(P_PAPERS, p)) {
		fprintf(stderr,
			"PAPERS PE:  need %s hardware\n",
			P_PAPERS);
		exit(1);
	}
	p = getenv("PAPERS_NPROC");
	if ((p != NULL) && (NPROC != atoi(p))) {
		fprintf(stderr,
			"PAPERS PE:  need %d PEs, environment has %d\n",
			NPROC, atoi(p));
		exit(1);
	}
	p = getenv("PAPERS_PORTBASE");
	if ((p != NULL) && (P_PORTBASE != atoi(p))) {
		fprintf(stderr,
			"PAPERS PE:  compiled for port at %d, environment has %d\n",
			P_PORTBASE, atoi(p));
		exit(1);
	}

	p = getenv("PAPERS_IPROC");
	iproc = ((p == NULL) ? -1 : atoi(p));
	switch (iproc) {
	case 0:
	case 1:
	case 2:
	case 3:
		break;
	default:
		fprintf(stderr,
			"PAPERS PE:  illegal PAPERS PE number (%d)\n",
			iproc);
		exit(1);
	}

	/* Get permission to access the port */
	P_PERM();
}

void
p_init(void)
{
	/* PAPERS initialization routine...
	   this ****ASSUMES**** that a proper p_exit() call
	   was used to terminate the previous parallel code.
	*/

	/* First, initialize environment info... */
	p_initenv();

	/* Do a simple sync sequence to ensure that all are here... */
	p_hw_startup();
}

void
p_error(register char *s)
{
	/* Print error message and die... */
	fprintf(stderr,
		"\nPAPERS PE%d:  %s\n",
		IPROC,
		s);

	/* Cause everybody to die... */
	p_exit(1);
	exit(1);
}

void
p_exit(register int ecode)
{
	/* Exit happened...  send everyone an interrupt and set our data bit
	   to indicate if we had a non-zero error code.  Well, we can't
	   quite do that... yet.  Just set S bits so that we will sync with
	   anything, turn our LED off, and send a data value that
	   effectively makes us appear not to be present.  Note: this does
	   not actually exit the user program.
	*/

	P_OUT(P_S | P_LB | ALL_MASK);
}

/*	The following code is for handling a signal from the PAPERS unit.
	We don't actually know who caused the interrupt, but by default we
	just die...  so it doesn't really matter who caused the problem.
	Alternatively, the signal could be used to transfer execution to a
	different program segment and then to send other info about the
	cause of the interrupt and the action desired.  In any case, the
	signal does not even have to interfere with normal PAPERS operations
	in progress, since the return value from the signal handler is used
	as the return value for the P_IN() that caught the interrupt....
*/

unsigned char
p_die(unsigned char x)
{
	fprintf(stderr,
		"PAPERS interrupt detected -- parallel job killed\n");

	/* Indicate that parallel program terminated abnormally */
	p_exit(1);
	exit(1);

	/* To make GCC happy... */
	return(x);
}

unsigned char (*p_signalHandler)(unsigned char) = p_die;

void
p_signal(unsigned char (*fnptr)(unsigned char))
{
	p_signalHandler = fnptr;
}
